//This program is licensed under the GPLv3; see license.txt for more info.
//Copyright 2007 Sprite_tm

#include "io.h"
#include "avr/interrupt.h"
#include "util/delay.h"
#include "lcd.h"
#include <string.h>
#include <avr/pgmspace.h>
#include "sintab.h"
#include "anim.h"


#define CONTRAST 107
#define CONTRAST_BLACK 130
#define CONTRAST_WHITE 0

//Send a string in progmem to the LCD
static void lcdstr(char p, char *c) {
    char t=0;
    char i;
    lcdcmd(0x80+p); //write to ddram, position p
    //write bytes until a 0 is found
    while ((i=pgm_read_byte(c+t))!=0) {
	lcdchar(i);
	t++;
    }
}

//Fade the contrast from one value (f) to another (t).
static void fade_to(unsigned char f, unsigned char t) {
    while(f!=t) {
	lcd_contrast(f);
	_delay_ms(40);
	if (f>t) f--; else f++;
    }
}

//Waits a certain time. Why? _delay_ms seems to hiccup on larger values :/
static void wait(int ms) {
    while (ms>10) {
	_delay_ms(10);
	ms-=10;
    }
//Don't uncomment this, if you do you'll lose 1K on floating point libs.
//    _delay_ms(ms);
}


//Looks up a sine value (p=0-255 instead of 0-2*pi) and returns
//the result, in a range from -128 to 127.
//This routine uses only a quarter of a sine table, because
//sin(0 to .5pi) = sin(pi to .5pi) and
//sin(0 to -pi) = -sin(0 to pi).
const static char mysin(const char p) {
#if (1)
//In theory quicker, because non-re-entrancy, in practice slower
//and using more memory, it seems.
    if (p==-128) return 0;

    if (p<0) {
	if (p<-63) {
	    return -pgm_read_byte(sintab+(p+127));
	} else {
	    return -pgm_read_byte(sintab-p);
	}
    } else {
	if (p<63) {
	    return pgm_read_byte(sintab+p);
	} else {
	    return pgm_read_byte(sintab+(127-p));
	}
    }

#else
//Recursive but using less program memory.
    if (p==-128) return mysin(0);
    if (p<-63) return -mysin(127+p);
    if (p<0) return -mysin(-p);
    if (p>63) return mysin(127-p);
#endif
    
}



const char banner[] PROGMEM="     ...can be abused to do some gfx-work!     ";

//This routine sends a bunch of custom chars stored in progmem.
//pos is the location of the chars, offset is the starting custom char
//position in HD44780 CGRAM, n is the amount of chars.

//Characters are stored a vertical line per byte because we only need 5 bytes
//per character when using that, but the HD44780 needs them as 8 bytes
//each containing a horizontal line.
static void send_custom_chars(unsigned char *pos, char offset, char n) {
    char x,y,t,ch;
    char p1,p2,p3,p4,p5,p6,p7,p8;
    lcdcmd(0x40+(offset<<3));
    for (ch=0; ch<n; ch++) {
	//Convert the chars from horizontal-line-per-byte to vertical-line-per-byte
	for (x=0; x<5; x++) {
	    y=pgm_read_byte(pos++);
	    p1<<=1; if (y&1) p1++;
	    p2<<=1; if (y&2) p2++;
	    p3<<=1; if (y&4) p3++;
	    p4<<=1; if (y&8) p4++;
	    p5<<=1; if (y&16) p5++;
	    p6<<=1; if (y&32) p6++;
	    p7<<=1; if (y&64) p7++;
	    p8<<=1; if (y&128) p8++;
	}
	//send the chars
	lcdchar(p8);
	lcdchar(p7);
	lcdchar(p6);
	lcdchar(p5);
	lcdchar(p4);
	lcdchar(p3);
	lcdchar(p2);
	lcdchar(p1);
    }
}

//Effect: Show the banner with the nice sitting female on the left
//Does a fade-in on start.
static void do_female_banner(void) {
    unsigned char x,bpos=0;
    unsigned short n;
    unsigned char contr=CONTRAST_BLACK;
    //Set up custom characters for the female
    lcdcmd(1);
    lcdcmd(0x80);
    lcdchar(0);
    lcdchar(2);
    lcdchar(4);
    lcdcmd(0x80+0x40);
    lcdchar(1);
    lcdchar(3);
    lcdchar(5);
    lcdchar(6);
    lcdchar(7);

    for (n=0; n<38*8; n++) {
	//fade in
	if (contr>CONTRAST) contr--;
	lcd_contrast(contr);

	//Each round through the loop, send an alternating image
	//for the grayscale-effect.
	if (n&1) {
	    send_custom_chars(custom_chars+CUST_GIRL_RESC0, 0, 8);
	} else {
	    send_custom_chars(custom_chars+CUST_GIRL_RESC1, 0, 8);
	}

	//re-send the banner every 8 iterations.
	if ((n&7)==0) {
	    lcdcmd(0x80+3);
	    for (x=0; x<5; x++) {
		lcdchar(pgm_read_byte(banner+x+bpos));
	    }
	    bpos++;
	}
    }
}


//Gets the value of a pixel in plasma.
//Used by do_plasma
const static char getpixel(const char x, const char y, const short n) {
    char p;

    //This is the base of the plasma: nothing more than some strange
    //random sine functions. Modify these to modify the shape of the plasma.
    p=mysin(x+n)/4;
    p+=mysin((x<<2)+(y<<1)+(n<<2))/4;
    p+=mysin(-(x<<3)+(y<<1)+(n>>1))/2;
    p+=mysin((x<<1)-(y<<3)+(n<<1))/2;
    p+=mysin((x<<4)+(y<<4)+(n>>3))/4;
    
    //Take the sine of the result and shift to 2 bits.
    p=(mysin(p<<1)+128)>>6;

    if ((p&2) && (p&1)) return 1;
    //For grayscale:
    if (((p&2)||(p&1)) && ((n+x+y)&1)) return 1;

//    if (mysin((x)<<4)>(y<<4)) return 1;
    return 0;
}

//Do plasma
static void do_plasma() {
    char x,t,y,xx,dx=0,dy=0,col;
    short n;
    //set up field of custom characters
    lcdcmd(1);
    lcdcmd(0x82);
    for (x=0; x<4; x++) lcdchar(x<<1);
    lcdcmd(0x80+0x42);
    for (x=0; x<4; x++) lcdchar((x<<1)+1);

    while(1) {
	//Send pixel data as custom character defines
	lcdcmd(0x40);
	dx=0;
        for (x=0; x<20; x+=5) {
	    dy=0;
	    for (y=0; y<16; y++) {
		for (xx=0; xx<5; xx++) {
		    t=t<<1;
		    if (getpixel(x+xx+dx,y+dy,n)) t++;
		}
		lcdchar(t);
		if (y==7) dy++;
	    }
	    dx++;
	}
	n++;
    }

}
//Send dancing girl over to lcd
static void girldance(void) {
    unsigned short n,noff;
    //set up field
    lcdcmd(0x80);
    lcdchar(0);
    lcdchar(2);
    lcdstr(2,PSTR("ANI-"));
    lcdchar(0);
    lcdchar(2);
    lcdcmd(0x80+0x40);
    lcdchar(1);
    lcdchar(3);
    lcdstr(0x42,PSTR("MATE"));
    lcdchar(1);
    lcdchar(3);
    noff=0;
    //Animate the girl!
    for (n=0; n<250; n++) {
	send_custom_chars(custom_chars+ANIM_GIRL+noff, 0, 4);
	if ((n&1)==0) noff+=(4*5); //every 2 iterations, use the next animation frame.
	if (noff==4*5*30) noff=0; //30 frames of 4 chars of 5 bytes.
    }
}


//L4m3 text intro :)
static void do_lame_intro() {
    lcd_contrast(255);
    lcdstr(0,   PSTR("HD44780 "));
    lcdstr(0x40,PSTR(" can be "));
    wait(500);
    lcd_contrast(CONTRAST);
    wait(2000);
    lcdstr(0,   PSTR("used to "));
    lcdstr(0x40,PSTR(" display"));
    wait(2000);
    lcdstr(0,   PSTR("  some  "));
    lcdstr(0x40,PSTR("text, or"));
    wait(1000);
    fade_to(CONTRAST,CONTRAST_BLACK);
}

int main(void) {
    //Startup delay
//    _delay_ms(50);
    //Init IO
    initio();
    //Init LCD
    initlcd();
    lcd_contrast(CONTRAST);
    lcd_backlight(128);

    do_lame_intro();
    do_female_banner();
    girldance();
    do_plasma();    

}
